package io.tswf.groovy.bettergroovy.io.tree.provider

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.io.tree.IFileTree
import io.tswf.groovy.bettergroovy.io.tree.IFileTreeEntry
import io.tswf.groovy.bettergroovy.io.tree.IFileTreeProvider

import java.nio.file.FileSystem
import java.nio.file.FileSystems
import java.nio.file.Files
import java.nio.file.NoSuchFileException
import java.nio.file.Path
import java.nio.file.Paths
import java.util.stream.Stream

@AutoService(IFileTreeProvider)
@CompileStatic
class NioFileSystemTreeProvider implements IFileTreeProvider {

    public static final Set<String> AVAILABLE_EXTENSIONS = [ "zip", "jar" ] as Set

    @Override
    IFileTree getFileTree(Object source) {
        def convertedSource = convertToPath(source)
        def fs
        def root = null
        def closable = true

        if (Files.isDirectory(convertedSource)) {
            fs = convertedSource.fileSystem
            root = convertedSource
            closable = false
        } else {
            fs = FileSystems.newFileSystem(convertedSource, (ClassLoader) null)
        }

        if (fs == null) {
            throw new RuntimeException("Can't create filesystem for source object '${source}' (converted to path: '${convertedSource}')!")
        }

        return new NioFileSystemTreeAdapter(fs, root, closable)
    }

    @Override
    boolean acceptsRawSource(Object source) {
        if (source instanceof File) {
            if (Files.isDirectory(source.absoluteCanonicalFile.toPath())) {
                return true
            } else {
                if (source.nameExtension in AVAILABLE_EXTENSIONS) {
                    return true
                }
            }
        }
        return false
    }

    Path convertToPath(Object source) {
        if (!acceptsRawSource(source)) {
            throw new IllegalArgumentException("The source '${source}' with class ${source.class.name} is not supported by ${this}!")
        }

        return ((File) source).toPath()
    }
}

@CompileStatic
class NioFileSystemTreeAdapter implements IFileTree {

    boolean closable
    FileSystem wrappedFileSystem
    Path root

    NioFileSystemTreeAdapter(FileSystem wrappedFileSystem, Path root, boolean closable) {
        this.wrappedFileSystem = wrappedFileSystem
        this.root = root
        this.closable = closable
    }

    @Override
    IFileTreeEntry getEntry(String name) {
        def path = getPathUsingRoot(name)
        if (Files.exists(path)) {
            return new NioPathTreeEntryAdapter(path, root)
        }

        return null
    }

    @Override
    Stream<IFileTreeEntry> getEntries() {
        return getEntries("")
    }

    @Override
    Stream<IFileTreeEntry> getEntries(String path) {
        def streamToWalk

        try {
            streamToWalk = Files.walk(getPathUsingRoot(path))
        } catch (NoSuchFileException e) {
            return Stream.<IFileTreeEntry>empty()
        }

        return streamToWalk
                .map {(IFileTreeEntry) new NioPathTreeEntryAdapter(it, root)  }
    }

    void close() throws IOException {
        if (closable)
            wrappedFileSystem.close()
    }

    Path getPathUsingRoot(String path) {
        if (root == null) {
            return wrappedFileSystem.getPath(path)
        } else {
            return wrappedFileSystem.getPath(root.toString(), path.toString())
        }
    }
}

@CompileStatic
class NioPathTreeEntryAdapter implements IFileTreeEntry {

    Path relativePath
    Path wrappedPath
    Path root

    NioPathTreeEntryAdapter(Path wrappedPath, Path root) {
        this.wrappedPath = wrappedPath
        this.root = root
        this.relativePath = root == null ? wrappedPath : root.relativize(wrappedPath)
    }

    @Override
    String getName() {
        return relativePath.toString().replace("\\", "/")
    }

    @Override
    InputStream getInputStream() {
        return Files.newInputStream(wrappedPath)
    }
}