package io.tswf.groovy.bettergroovy.scan

import groovy.transform.CompileStatic

import java.util.concurrent.ConcurrentHashMap
import java.util.function.Supplier
import java.util.zip.ZipEntry
import java.util.zip.ZipFile
import java.util.zip.ZipInputStream

@CompileStatic
abstract class AbstractZipFileProvider implements IResourcesProvider {

    Map<Object, Set<String>> entriesCache = new ConcurrentHashMap<>()

    @Override
    void getResources(ResourceScanContext context) {
        if (!tryResolveByZipFile(context)) {
            resolveByZipInputStream(context)
        }
    }

    boolean tryResolveByZipFile(ResourceScanContext context) {
        def resourceNameStart = context.packageName?.replace(".", "/") ?: ""
        def zipFile = getZipFile(context.source)
        if (zipFile == null)
            return false

        def entries = zipFile.stream()
            .filter {it.name.startsWith(resourceNameStart)}
            .map {getResourceFromZipEntry(context.source, zipFile, it)}
            .filter {context.filter == null || context.filter.test(it)}

        context.foundResources += entries

        return true
    }

    void resolveByZipInputStream(ResourceScanContext context) {
        def zipIs = getZipFileInputStream(context.source)
        def resourceNameStart = context.packageName?.replace(".", "/") ?: ""
        def sourceObject = context.source
        def cacheKey = sourceObject

        if (sourceObject instanceof URL) {
            cacheKey = sourceObject.toString()
        } else if (sourceObject instanceof File) {
            cacheKey = sourceObject.absoluteCanonicalFile.path
        }

        def entriesList = entriesCache.computeIfAbsent(cacheKey) {
            zipIs.entriesStream()
                    .map {it.name}
                    .toSet()
        }

        def packageMatchingEntriesList = entriesList.filter {it.startsWith(resourceNameStart)}
        if (packageMatchingEntriesList.isEmpty()) {
            return
        }

        context.foundResources += findAllMatchingResources(packageMatchingEntriesList, context).stream()
    }

    List<Resource> findAllMatchingResources(Collection<String> names, ResourceScanContext context) {
        return getZipFileInputStream(context.source).doThenClose {zis ->
            def result = new ArrayList<Resource>()
            ZipEntry entry
            while ((entry = zis.nextEntry) != null) {
                if (entry.name !in names)
                    continue

                def resource = getResourceFromStreamZipEntry(context.source, zis, entry)

                if (context.filter == null || context.filter.test(resource)) {
                    result << resource
                }
            }

            return result
        }
    }

    @Override
    boolean acceptsContext(ResourceScanContext context) {
        return acceptsSource(context.source)
    }

    @Override
    boolean acceptsContext(ResourceGetContext context) {
        return acceptsSource(context.source)
    }

    boolean acceptsSource(Object source) {
        return tryZipFile(source)
    }

    boolean tryZipFile(Object source) {
        try {
            return getZipFileInputStream(source)?.entriesStream() != null
        } catch (Throwable skip){
            return false
        }
    }

    Resource getResourceFromStreamZipEntry(Object owner, ZipInputStream zis, ZipEntry entry) {
        def resourceBytes = zis.readZipEntryContent()
        def resource = new Resource(
                name: entry.name,
                fullPath: entry.name,
                owner: owner,
                inputStreamSupplier: {
                    new ByteArrayInputStream(resourceBytes)
                }
        )

        return resource
    }

    Resource getResourceFromZipEntry(Object owner, ZipFile zipFile, ZipEntry zipEntry) {
        return new Resource(
                name: zipEntry.name,
                fullPath: zipEntry.name,
                owner: owner,
                inputStreamSupplier: {
                    return zipFile.getInputStream(zipEntry)
                } as Supplier<InputStream>
        )
    }

    @Override
    void getResource(ResourceGetContext context) {
        def locations = context.locations
        if (locations.isNotNullOrEmpty()) {
            return
        }

        // Пробуем резолв через ZipFile
        def zipFileFoundResource = getZipFile(context)?.with {zipFile ->
            def resourcesFoundForZipFile = new ArrayList<Resource>()
            for (def location : context.locations) {
                location = location.stripAll()

                if (zipFile != null) {
                    def foundEntry = zipFile.getEntry(location)
                    if (foundEntry != null) {
                        def res = context.with { getResourceFromZipEntry(source, zipFile, foundEntry) }
                        if (res != null) {
                            resourcesFoundForZipFile << res
                        }
                    }
                }
            }

            return resourcesFoundForZipFile
        }

        if (zipFileFoundResource.isNotNullOrEmpty()) {
            context.foundResource.addAll(zipFileFoundResource)
            return
        }

        context.foundResource.addAll getZipFileInputStream(context.source)?.doThenClose { zipInputStream ->
            def resourcesFoundForZis = new ArrayList<Resource>()
            def locationsToSearch = locations.toSet()

            def entry = (ZipEntry) null
            while (entry = zipInputStream.nextEntry) {
                if (entry.name in locationsToSearch) {
                    def res = getResourceFromStreamZipEntry(context.source, zipInputStream, entry)

                    if (res != null) {
                        locationsToSearch.remove(entry.name)
                        resourcesFoundForZis << res
                    }
                }

                if (locationsToSearch.isEmpty()) {
                    break
                }
            }

            return resourcesFoundForZis
        }
    }

    abstract ZipInputStream getZipFileInputStream(Object source)
    abstract ZipFile getZipFile(Object source)
}
