package io.tswf.groovy.bettergroovy.scan

import groovy.transform.CompileStatic
import groovy.transform.ToString
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor

import java.util.function.Predicate
import java.util.function.Supplier
import java.util.stream.Stream

@CompileStatic
@Singleton(lazy = true)
class ClasspathResourceScanner {

    List<IResourcesProvider> resourcesProviders = ServiceLoader.load(IResourcesProvider).toList()
    List<?> classpath = {
        def classpath = System.getClasspathList()

        return classpath.map {url ->
            def protocol = url.protocol
            if (protocol == "file") {
                return new File(url.path)
            }

            return url
        }
    }()

    Collection<Resource> getResources(
            Iterable<String> locations,
            Predicate<Resource> filter = { true },
            Iterable<Object> sources = classpath
    ) {
        def result = new ArrayList<Resource>()

        for (def classpathEntry : sources) {
            try {
                def getContext = new ResourceGetContext(
                        filter: filter,
                        source: classpathEntry,
                        locations: locations.toSet()
                )

                for (def resourceProvider : resourcesProviders) {
                    if (resourceProvider.acceptsContext(getContext))
                        resourceProvider.getResource(getContext)
                }

                getContext.foundResource?.let {
                    if (it.isNotNullOrEmpty())
                        result.addAll(it)
                }
            } catch (Throwable e) {
                throw new ClasspathScanException("Error occurred while scanning '${classpathEntry}'", e)
            }
        }

        return result
    }

    Stream<Resource> findResourcesInPackage(
            String packageName,
            Predicate<Resource> filter = { true },
            Predicate<Object> sourceFilter = { true },
            Iterable<Object> sources = classpath
    ) {
        if (sources == null) {
            sources = classpath
        }

        def result = Stream.<Resource>empty()

        for (def classpathSource : sources) {
            if (!sourceFilter(classpathSource))
                continue

            try {
                def scanContext = new ResourceScanContext(
                        filter: filter,
                        packageName: packageName,
                        source: classpathSource
                )

                for (def resourceProvider : resourcesProviders) {
                    if (resourceProvider.acceptsContext(scanContext))
                        resourceProvider.getResources(scanContext)
                }

                result += scanContext.foundResources
            } catch (Throwable e) {
                throw new ClasspathScanException("Error occurred while scanning '${classpathSource}'", e)
            }
        }

        return result
    }
}

@CompileStatic
@CheckedMapConstructor
class ResourceScanContext {
    Object source
    Predicate<Resource> filter
    String packageName
    Stream<Resource> foundResources = new ArrayList<Resource>().stream()
}

@CompileStatic
@CheckedMapConstructor
class ResourceGetContext {
    Object source
    Set<String> locations
    Predicate<Resource> filter
    Set<Resource> foundResource = new HashSet<>()
}

@ToString(includes = ["owner", "fullPath"])
@CompileStatic
@CheckedMapConstructor
class Resource {
    String name
    String fullPath
    Supplier<InputStream> inputStreamSupplier
    Object owner
}

trait IResourcesProvider {
    abstract void getResource(ResourceGetContext context)
    abstract void getResources(ResourceScanContext context)
    abstract boolean acceptsContext(ResourceScanContext context)
    abstract boolean acceptsContext(ResourceGetContext context)
}
